@@ -1,5 +1,5 @@
 # -*-cperl-*-
-#  $Id: Pg.pm,v 1.153 2005/04/06 22:39:05 turnstep Exp $
+#  $Id: Pg.pm,v 1.159 2005/05/21 14:36:05 turnstep Exp $
 #
 #  Copyright (c) 2002-2005 PostgreSQL Global Development Group
 #  Portions Copyright (c) 2002 Jeffrey W. Baker
@@ -12,10 +12,11 @@
 
 use 5.006001;
 
-{
-	package DBD::Pg;
 
-	our $VERSION = '1.41';
+
+{ package DBD::Pg;
+
+	our $VERSION = '1.42';
 
 	use DBI ();
 	use DynaLoader ();
@@ -23,8 +24,9 @@ use 5.006001;
 	use vars qw(@ISA %EXPORT_TAGS $err $errstr $sqlstate $drh $dbh);
 	@ISA = qw(DynaLoader Exporter);
 
-	%EXPORT_TAGS = (
-		pg_types => [qw(
+	%EXPORT_TAGS = 
+		(
+		 pg_types => [qw(
 			PG_BOOL PG_BYTEA PG_CHAR PG_INT8 PG_INT2 PG_INT4 PG_TEXT PG_OID
 			PG_FLOAT4 PG_FLOAT8 PG_ABSTIME PG_RELTIME PG_TINTERVAL PG_BPCHAR
 			PG_VARCHAR PG_DATE PG_TIME PG_DATETIME PG_TIMESPAN PG_TIMESTAMP
@@ -37,16 +39,17 @@ use 5.006001;
 
 	bootstrap DBD::Pg $VERSION;
 
-	$err = 0;		# holds error code for DBI::err
-	$errstr = "";	# holds error string for DBI::errstr
-	$sqlstate = "";
-	$drh = undef;	# holds driver handle once initialized
+	$err = 0;		    # holds error code for DBI::err
+	$errstr = "";	  # holds error string for DBI::errstr
+	$sqlstate = ""; # holds five character SQLSTATE code
+	$drh = undef;	  # holds driver handle once initialized
 
 	sub CLONE {
 		$drh = undef;
 	}
 
-	sub driver{
+
+	sub driver {
 		return $drh if defined $drh;
 		my($class, $attr) = @_;
 
@@ -77,6 +80,7 @@ use 5.006001;
 
 	}
 
+
  	## Deprecated: use $dbh->{pg_server_version} if possible instead
  	sub _pg_use_catalog {
 		my $dbh = shift;
@@ -89,12 +93,14 @@ use 5.006001;
 }
 
 
-{ package DBD::Pg::dr; # ====== DRIVER ======
+{ package DBD::Pg::dr;
 
 	use strict;
 
-	our $CATALOG = 123;
+	our $CATALOG = 123; ## Set later on, this is to catch seriously misplaced code
 
+
+	## Returns an array of formatted database names from the pg_database table
 	sub data_sources {
 		my $drh = shift;
 		my $dbh = DBD::Pg::dr::connect($drh, 'dbname=template1') or return undef;
@@ -109,9 +115,7 @@ use 5.006001;
 
 
 	sub connect {
-		my($drh, $dbname, $user, $auth)= @_;
-
-		# create a 'blank' dbh
+		my($drh, $dbname, $user, $pass)= @_;
 
 		## Allow "db" and "database" as synonyms for "dbname"
 		$dbname =~ s/\b(?:db|database)\s*=/dbname=/;
@@ -125,23 +129,22 @@ use 5.006001;
 			$Name = $1;
 		}
 	
-	
-		$user = "" unless defined($user);
-		$auth = "" unless defined($auth);
+ 		$user = "" unless defined($user);
+		$pass = "" unless defined($pass);
 
 		$user = $ENV{DBI_USER} if $user eq "";
-		$auth = $ENV{DBI_PASS} if $auth eq "";
+		$pass = $ENV{DBI_PASS} if $pass eq "";
 
 		$user = "" unless defined($user);
-		$auth = "" unless defined($auth);
+		$pass = "" unless defined($pass);
 
-		my($dbh) = DBI::_new_dbh($drh, {
+		my ($dbh) = DBI::_new_dbh($drh, {
 			'Name' => $Name,
 			'User' => $user, 'CURRENT_USER' => $user,
 		});
 
 		# Connect to the database..
-		DBD::Pg::db::_login($dbh, $dbname, $user, $auth) or return undef;
+		DBD::Pg::db::_login($dbh, $dbname, $user, $pass) or return undef;
 
 		my $version = $dbh->{pg_server_version};
 		$dbh->{private_dbdpg}{version} = $version;
@@ -155,16 +158,15 @@ use 5.006001;
 }
 
 
-{ package DBD::Pg::db; # ====== DATABASE ======
+{ package DBD::Pg::db;
 
 	use strict;
-	use Carp ();
+
 
 	sub prepare {
-		my($dbh, $statement, @attribs)= @_;
-
-		# create a 'blank' sth
+		my($dbh, $statement, @attribs) = @_;
 
+		# Create a 'blank' statement handle:
 		my $sth = DBI::_new_sth($dbh, {
 			'Statement' => $statement,
 		});
@@ -187,6 +189,7 @@ use 5.006001;
 		$sth;
 	}
 
+
 	sub do {
 		my ($dbh, $statement, $attr, @params) = @_;
 
@@ -205,6 +208,7 @@ use 5.006001;
 		($rows == 0) ? "0E0" : $rows;
 	}
 
+
 	sub last_insert_id {
 
 		my ($dbh, $catalog, $schema, $table, $col, $attr) = @_;
@@ -226,7 +230,7 @@ use 5.006001;
 				$attr = {sequence => $attr};
 			}
 			elsif (ref $attr ne 'HASH') {
-				die "last_insert_id must be passed a hashref as the final argument\n";
+				return $dbh->set_err(1, "last_insert_id must be passed a hashref as the final argument");
 			}
 			## Named sequence overrides any table or schema settings
 			if (exists $attr->{sequence} and length $attr->{sequence}) {
@@ -243,7 +247,7 @@ use 5.006001;
 		elsif (! defined $sequence) {
 			## At this point, we must have a valid table name
 			if (! length $table) {
-				die qq{last_insert_id needs at least a sequence or table name\n};
+				return $dbh->set_err(1, "last_insert_id needs at least a sequence or table name");
 			}
 			my @args = ($table);
 
@@ -264,7 +268,7 @@ use 5.006001;
 				$sth->finish();
 				my $message = qq{Could not find the table "$table"};
 				length $schema and $message .= qq{ in the schema "$schema"};
-				die "$message\n";
+				return $dbh->set_err(1, $message);
 			}
 			my $oid = $sth->fetchall_arrayref()->[0][0];
 			## This table has a primary key. Is there a sequence associated with it via a unique, indexed column?
@@ -278,7 +282,7 @@ use 5.006001;
 			$count = $sth->execute();
 			if (!defined $count or $count eq '0E0') {
 				$sth->finish();
-				die qq{No suitable column found for last_insert_id of table "$table"\n};
+				$dbh->set_err(1, qq{No suitable column found for last_insert_id of table "$table"});
 			}
 			my $info = $sth->fetchall_arrayref();
 
@@ -290,13 +294,13 @@ use 5.006001;
 				push @def, $_;
 			}
 			if (!@def) {
-				die qq{No suitable column found for last_insert_id of table "$table"\n};
+				$dbh->set_err(1, qq{No suitable column found for last_insert_id of table "$table"\n});
 			}
 			## Tiebreaker goes to the primary keys
 			if (@def > 1) {
 				my @pri = grep { $_->[1] } @def;
 				if (1 != @pri) {
-					die qq{No suitable column found for last_insert_id of table "$table"\n};
+					$dbh->set_Err(1, qq{No suitable column found for last_insert_id of table "$table"\n});
 				}
 				@def = @pri;
 			}
@@ -329,7 +333,7 @@ use 5.006001;
 
 	# Column expected in statement handle returned.
 	# table_cat, table_schem, table_name, column_name, data_type, type_name,
-	# column_size, buffer_length, DECIMAL_DIGITS, NUM_PREC_RADIX, NULLABLE,
+ 	# column_size, buffer_length, DECIMAL_DIGITS, NUM_PREC_RADIX, NULLABLE,
 	# REMARKS, COLUMN_DEF, SQL_DATA_TYPE, SQL_DATETIME_SUB, CHAR_OCTET_LENGTH,
 	# ORDINAL_POSITION, IS_NULLABLE
 	# The result set is ordered by TABLE_SCHEM, TABLE_NAME and ORDINAL_POSITION.
@@ -363,7 +367,7 @@ use 5.006001;
 			"JOIN pg_catalog.pg_namespace n ON (n.oid = c.relnamespace)" : "";
 
 		# col_description is not available for Pg < 7.2
-		my $remarks = $version < 70200 ? 
+		my $remarks = $version > 70200 ? 
 			"${DBD::Pg::dr::CATALOG}col_description(a.attrelid, a.attnum)" : "NULL::text";
 
 		my $col_info_sql = qq!
@@ -1307,14 +1311,12 @@ use 5.006001;
 		 }
 		 return $ans;
 	} # end of get_info
-} # end of package DBD::Pg::db
-
-{ package DBD::Pg::st; # ====== STATEMENT ======
+} 
 
 
-	## The DBI version is broken, so we implement a near-copy here
+{ package DBD::Pg::st;
 
-	sub bind_param_array {
+	sub bind_param_array { ## The DBI version is broken, so we implement a near-copy here
 		my $sth = shift;
 		my ($p_id, $value_array, $attr) = @_;
 
@@ -2736,7 +2738,7 @@ $dbh->func($data, 100, 'getline').
 
 This driver supports all largeobject functions provided by libpq via the
 C<func> method. Please note that, starting with PostgreSQL 6.5, any access to
-a large object -- even read-only large objects -- must to be put into a
+a large object -- even read-only large objects -- must be put into a
 transaction!
 
 =head2 Cursors
@@ -2793,7 +2795,7 @@ DBD-Pg by Edmund Mergl (E.Mergl@bawue.de) and Jeffrey W. Baker
 Stewart <jason@openinformatics.com> and Bruce Momjian
 <pgman@candle.pha.pa.us> and others after v1.13.
 
-Major parts of this package have been copied from DBI and DBD-Oracle.
+Parts of this package have been copied from DBI and DBD-Oracle.
 
 B<Mailing List>
 
