@@ -1,6 +1,6 @@
 /*
 
-   $Id: dbdimp.c,v 1.116 2005/04/05 13:48:55 turnstep Exp $
+   $Id: dbdimp.c,v 1.131 2005/05/18 18:56:01 turnstep Exp $
 
    Copyright (c) 2002-2005 PostgreSQL Global Development Group
    Portions Copyright (c) 2002 Jeffrey W. Baker
@@ -75,8 +75,10 @@ void dbd_st_split_statement();
 int dbd_st_prepare_statement();
 int dbd_db_transaction_status();
 int dbd_st_deallocate_statement();
+int dbd_db_rollback_commit();
+int is_high_bit_set();
 PGTransactionStatusType dbd_db_txn_status();
-#include "large_object.c"
+
 
 /* ================================================================== */
 /* Quick result grabber used throughout this file */
@@ -85,11 +87,13 @@ ExecStatusType _result(imp_dbh, com)
 		 const char *com;
 {
 	PGresult *result;
-	ExecStatusType status;
+	int status = -1;
+
+	if (dbis->debug >= 1) PerlIO_printf(DBILOGFP, "Running _result with (%s)\n", com);
 
 	result = PQexec(imp_dbh->conn, com);
-
-	status = result ? PQresultStatus(result) : -1;
+	if (result)
+		status = PQresultStatus(result);
 
 #if PGLIBVERSION >= 70400
 	strncpy(imp_dbh->sqlstate,
@@ -267,11 +271,10 @@ int dbd_db_login (dbh, imp_dbh, dbname, uid, pwd)
 	imp_dbh->pg_server_version = -1;
 	{
 		PGresult *result;
-		ExecStatusType status;
-		int	cnt, vmaj, vmin, vrev;
+		int	status, cnt, vmaj, vmin, vrev;
 	
 		result = PQexec(imp_dbh->conn, "SELECT version(), 'DBD::Pg'");
-		status = result ? PQresultStatus(result) : -1;
+		status = result ? (int)PQresultStatus(result) : -1;
 	
 		if (PGRES_TUPLES_OK != status || !PQntuples(result)) {
 			if (dbis->debug >= 4)
@@ -322,7 +325,7 @@ int dbd_db_ping (dbh)
 		 SV *dbh;
 {
 	D_imp_dbh(dbh);
-	ExecStatusType status;
+	int status;
 
 	/* Since this is a very explicit call, we do not rely on PQstatus,
 		 which can have stale information */
@@ -367,7 +370,7 @@ int dbd_db_rollback_commit (dbh, imp_dbh, action)
 {
 
 	PGTransactionStatusType tstatus;
-	ExecStatusType status;
+	int status;
 
 	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "%s\n", action); }
 	
@@ -401,7 +404,7 @@ int dbd_db_rollback_commit (dbh, imp_dbh, action)
 	}
 #endif
 
-	/* If begin_work has been called, turn AutoCommit back on and BeginWork off */
+	/* If begin_work has been called, turn AutoCommit back on and BegunWork off */
 	if (DBIc_has(imp_dbh, DBIcf_BegunWork)) {
 		DBIc_set(imp_dbh, DBIcf_AutoCommit, 1);
 		DBIc_set(imp_dbh, DBIcf_BegunWork, 0);
@@ -419,6 +422,10 @@ int dbd_db_rollback_commit (dbh, imp_dbh, action)
 
 	av_clear(imp_dbh->savepoints);
 	imp_dbh->done_begin = 0;
+
+	/* If we just did a rollback or a commit, we can no longer be in a PGRES_COPY state */
+	imp_dbh->copystate=0;
+
 	return 1;
 
 } /* end of dbd_db_rollback_commit */
@@ -501,16 +508,16 @@ int dbd_db_STORE_attrib (dbh, imp_dbh, keysv, valuesv)
 {
 	STRLEN kl;
 	char *key = SvPV(keysv,kl);
-	int oldval, newval = SvTRUE(valuesv);
+	int oldval;
+	int newval = SvTRUE(valuesv);
 
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_STORE\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_STORE (%s) (%d)\n", key, newval); }
 	
 	if (10==kl && strEQ(key, "AutoCommit")) {
 		oldval = DBIc_has(imp_dbh, DBIcf_AutoCommit);
 		if (oldval == newval)
 			return 1;
-		if (oldval) {
-			/* Commit if necessary */
+		if (newval) { /* It was off but is now on, so do a final commit */
 			if (dbd_db_commit(dbh, imp_dbh) && dbis->debug >= 5)
 				PerlIO_printf(DBILOGFP, "dbd_db_STORE: AutoCommit on forced a commit\n");
 		}
@@ -565,7 +572,7 @@ SV * dbd_db_FETCH_attrib (dbh, imp_dbh, keysv)
 	SV *retsv = Nullsv;
 	char *host = NULL;
 	
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_FETCH\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_FETCH: dbh=%s\n", dbh); }
 	
 	if (10==kl && strEQ(key, "AutoCommit")) {
 		retsv = boolSV(DBIc_has(imp_dbh, DBIcf_AutoCommit));
@@ -633,13 +640,12 @@ int dbd_discon_all (drh, imp_drh)
 {
 	dTHR;
 	
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_discon_all\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_discon_all: drh=%s\n", drh); }
 	
 	/* The disconnect_all concept is flawed and needs more work */
 	if (!PL_dirty && !SvTRUE(perl_get_sv("DBI::PERL_ENDING",0))) {
 		sv_setiv(DBIc_ERR(imp_drh), (IV)1);
-		sv_setpv(DBIc_ERRSTR(imp_drh),
-						 (char*)"disconnect_all not implemented");
+		sv_setpv(DBIc_ERRSTR(imp_drh), "disconnect_all not implemented");
 	}
 	return DBDPG_FALSE;
 
@@ -652,7 +658,7 @@ int dbd_db_getfd (dbh, imp_dbh)
 		 imp_dbh_t *imp_dbh;
 {
 
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_getfd\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_db_getfd: dbh=%s\n", dbh); }
 	
 	return PQsocket(imp_dbh->conn);
 
@@ -796,7 +802,7 @@ int dbd_st_prepare (sth, imp_sth, statement, attribs)
 	statement -= mypos; /* Rewind statement */
 
 	/* Break the statement into segments by placeholder */
-	dbd_st_split_statement(sth, imp_sth, statement);
+	dbd_st_split_statement(imp_sth, statement);
 
 	/*
 		We prepare it right away if:
@@ -830,10 +836,9 @@ int dbd_st_prepare (sth, imp_sth, statement, attribs)
 
 
 /* ================================================================== */
-void dbd_st_split_statement (sth, imp_sth, statement)
-		 SV *sth;
+void dbd_st_split_statement (imp_sth, statement)
 		 imp_sth_t *imp_sth;
-		 const char *statement;
+		 char *statement;
 {
 
 	/* Builds the "segment" and "placeholder" structures for a statement handle */
@@ -898,6 +903,14 @@ void dbd_st_split_statement (sth, imp_sth, statement)
 				if (!(backslashes & 1)) 
 					quote = 0;
 			}
+			else {
+				if ('\\' == ch) {
+					backslashes++;
+				}
+				else {
+					backslashes=0;
+				}
+			}
 			if (*statement)
 				continue;
 		}
@@ -1161,7 +1174,7 @@ int dbd_st_prepare_statement (sth, imp_sth)
 	unsigned int x;
 	STRLEN execsize;
 	PGresult *result;
-	ExecStatusType status;
+	int status = -1;
 	seg_t *currseg;
 	bool oldprepare = 1;
 	unsigned int params = 0;
@@ -1200,7 +1213,7 @@ int dbd_st_prepare_statement (sth, imp_sth)
 				continue;
 			/* The parameter itself: dollar sign plus digit(s) */
 			for (x=1; x<7; x++) {
-				if (currseg->placeholder < pow(10,x))
+				if (currseg->placeholder < pow((double)10,(double)x))
 					break;
 			}
 			if (x>=7)
@@ -1258,14 +1271,16 @@ int dbd_st_prepare_statement (sth, imp_sth)
 	else {
 		if (imp_sth->numbound) {
 			params = imp_sth->numphs;
-			Newz(0, paramTypes, imp_sth->numphs, Oid);
+			Newz(0, paramTypes, (unsigned)imp_sth->numphs, Oid);
 			for (x=0,currph=imp_sth->ph; NULL != currph; currph=currph->nextph) {
 				paramTypes[x++] = currph->defaultval ? 0 : currph->bind_type->type_id;
 			}
 		}
-		result = PQprepare(imp_dbh->conn, imp_sth->prepare_name, statement, params, paramTypes);
+		result = PQprepare(imp_dbh->conn, imp_sth->prepare_name, statement, (int)params, paramTypes);
 		Safefree(paramTypes);
-		status = result ? PQresultStatus(result) : -1;
+		if (result)
+			status = PQresultStatus(result);
+		PQclear(result);
 		if (dbis->debug >= 6)
 			PerlIO_printf(DBILOGFP, "  dbdpg: Using PQprepare\n");
 	}
@@ -1299,11 +1314,12 @@ int dbd_bind_ph (sth, imp_sth, ph_name, newvalue, sql_type, attribs, is_inout, m
 	char *name = Nullch;
 	STRLEN name_len;
 	ph_t *currph = NULL;
-	unsigned int x, phnum;
+	int x, phnum;
 	SV **svp;
 	bool reprepare = 0;
 	int pg_type = 0;
 	char *value_string = NULL;
+	maxlen = 0; /* not used */
 
 	if (dbis->debug >= 4) {
 		PerlIO_printf(DBILOGFP, "dbd_bind_ph\n");
@@ -1352,7 +1368,7 @@ int dbd_bind_ph (sth, imp_sth, ph_name, newvalue, sql_type, attribs, is_inout, m
 		if ('$' == *name)
 			*name++;
 		phnum = atoi(name);
-		if (phnum<1 || phnum>imp_sth->numphs)
+		if (phnum < 1 || phnum > imp_sth->numphs)
 			croak("Cannot bind unknown placeholder %d (%s)", phnum, neatsvpv(ph_name,0));
 		for (x=1,currph=imp_sth->ph; NULL != currph; currph = currph->nextph,x++) {
 			if (x==phnum)
@@ -1415,7 +1431,7 @@ int dbd_bind_ph (sth, imp_sth, ph_name, newvalue, sql_type, attribs, is_inout, m
  	}
 	else if (NULL == currph->bind_type) { /* "sticky" data type */
 		/* This is the default type, but we will honor defaultval if we can */
-		currph->bind_type = pg_type_data(VARCHAROID);
+				currph->bind_type = pg_type_data(UNKNOWNOID);
 		if (!currph->bind_type)
 			croak("Default type is bad!!!!???");
 	}
@@ -1479,13 +1495,13 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 
 	D_imp_dbh_from_sth;
 	ph_t *currph;
-	ExecStatusType status = -1;
+	int status = -1;
 	STRLEN execsize, x;
 	const char **paramValues = NULL;
 	int *paramLengths = NULL, *paramFormats = NULL;
 	Oid *paramTypes = NULL;
 	seg_t *currseg;
-	char *statement, *cmdStatus, *cmdTuples;
+	char *statement = NULL, *cmdStatus = NULL, *cmdTuples = NULL;
 	int num_fields, ret = -2;
 	
 	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_execute\n"); }
@@ -1544,6 +1560,8 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 				currph->quotedlen = 4;
 			}
 			else {
+				if (currph->quoted)
+					Safefree(currph->quoted);
 				currph->quoted = currph->bind_type->quote(currph->value, currph->valuelen, &currph->quotedlen);
 			}
 		}
@@ -1555,7 +1573,7 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 	}
 	else { /* We are using a server that can handle PQexecParams/PQexecPrepared */
 		/* Put all values into an array to pass to PQexecPrepared */
-		Newz(0, paramValues, imp_sth->numphs, const char *); /* freed below */
+		Newz(0, paramValues, (unsigned)imp_sth->numphs, const char *); /* freed below */
 		for (x=0,currph=imp_sth->ph; NULL != currph; currph=currph->nextph) {
 			paramValues[x++] = currph->value;
 		}
@@ -1563,8 +1581,8 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 		/* Binary or regular? */
 
 		if (imp_sth->has_binary) {
-			Newz(0, paramLengths, imp_sth->numphs, int); /* freed below */
-			Newz(0, paramFormats, imp_sth->numphs, int); /* freed below */
+			Newz(0, paramLengths, (unsigned)imp_sth->numphs, int); /* freed below */
+			Newz(0, paramFormats, (unsigned)imp_sth->numphs, int); /* freed below */
 			for (x=0,currph=imp_sth->ph; NULL != currph; currph=currph->nextph,x++) {
 				if (BYTEAOID==currph->bind_type->type_id) {
 					paramLengths[x] = currph->valuelen;
@@ -1610,6 +1628,9 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 					PerlIO_printf(DBILOGFP, "  dbdpg: re-preparing statement\n");
 			}
 			if (dbd_st_prepare_statement(sth, imp_sth)) {
+				Safefree(paramValues);
+				Safefree(paramLengths);
+				Safefree(paramFormats);
 				return -2;
 			}
 		}
@@ -1632,10 +1653,6 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 		imp_sth->result = PQexecPrepared(imp_dbh->conn, imp_sth->prepare_name, imp_sth->numphs,
 																		 paramValues, paramLengths, paramFormats, 0);
 
-		Safefree(paramValues);
-		Safefree(paramLengths);
-		Safefree(paramFormats);			
-
 	} /* end new-style prepare */
 	else {
 		
@@ -1658,7 +1675,7 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 					continue;
 				/* The parameter itself: dollar sign plus digit(s) */
 				for (x=1; x<7; x++) {
-					if (currseg->placeholder < pow(10,x))
+					if (currseg->placeholder < pow((double)10,(double)x))
 						break;
 				}
 				if (x>=7)
@@ -1679,7 +1696,7 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 			statement[execsize] = '\0';
 			
 			/* Populate paramTypes */
-			Newz(0, paramTypes, imp_sth->numphs, Oid);
+			Newz(0, paramTypes, (unsigned)imp_sth->numphs, Oid);
 			for (x=0,currph=imp_sth->ph; NULL != currph; currph=currph->nextph) {
 				paramTypes[x++] = currph->defaultval ? 0 : currph->bind_type->type_id;
 			}
@@ -1699,10 +1716,6 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 			imp_sth->result = PQexecParams(imp_dbh->conn, statement, imp_sth->numphs, paramTypes,
 																		 paramValues, paramLengths, paramFormats, 0);
 			Safefree(paramTypes);
-			Safefree(paramValues);
-			Safefree(paramLengths);
-			Safefree(paramFormats);
-			Safefree(statement);
 		}
 		
 		/* PQexec */
@@ -1732,15 +1745,23 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 				PerlIO_printf(DBILOGFP, "  dbdpg: calling PQexec for: %s\n", statement);
 			
 			imp_sth->result = PQexec(imp_dbh->conn, statement);
-			Safefree(statement);
 
 		} /* end PQexec */
 
+		Safefree(statement);
+
 	} /* end non-prepared exec */
 
+	Safefree(paramValues);
+	Safefree(paramLengths);
+	Safefree(paramFormats);			
+
 	/* Some form of PQexec has been run at this point */
 
-	status = imp_sth->result ? PQresultStatus(imp_sth->result) : -1;
+	if (imp_sth->result)
+		status = PQresultStatus(imp_sth->result);
+	else
+		status = -1;
 
 	/* We don't want the result cleared yet, so we don't use _result */
 
@@ -1754,9 +1775,11 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 	strcpy(imp_dbh->sqlstate, "S1000"); /* DBI standard says this is the default */
 #endif
 
-	cmdStatus = imp_sth->result ? (char *)PQcmdStatus(imp_sth->result) : "";
-	cmdTuples = imp_sth->result ? (char *)PQcmdTuples(imp_sth->result) : "";
-	
+	if (imp_sth->result) {
+		cmdStatus = PQcmdStatus(imp_sth->result);
+		cmdTuples = PQcmdTuples(imp_sth->result);
+	}
+
 	if (dbis->debug >= 5)
 		PerlIO_printf(DBILOGFP, "  dbdpg: received a status of %d\n", status);
 
@@ -1795,7 +1818,7 @@ int dbd_st_execute (sth, imp_sth) /* <= -2:error, >=0:ok row count, (-1=unknown
 	/* store the number of affected rows */
 	
 	imp_sth->rows = ret;
-	
+
 	return ret;
 
 } /* end of dbd_st_execute */
@@ -1849,9 +1872,9 @@ AV * dbd_st_fetch (sth, imp_sth)
 
 	/* Set up the type_info array if we have not seen it yet */
 	if (NULL==imp_sth->type_info) {
-		Newz(0, imp_sth->type_info, num_fields, sql_type_info_t*); /* freed in dbd_st_destroy */
+		Newz(0, imp_sth->type_info, (unsigned)num_fields, sql_type_info_t*); /* freed in dbd_st_destroy */
 		for (i = 0; i < num_fields; ++i) {
-			imp_sth->type_info[i] = pg_type_data(PQftype(imp_sth->result, i));
+			imp_sth->type_info[i] = pg_type_data((int)PQftype(imp_sth->result, i));
 		}
 	}
 	
@@ -1920,7 +1943,7 @@ int dbd_st_rows (sth, imp_sth)
 		 SV *sth;
 		 imp_sth_t *imp_sth;
 {
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_rows\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_rows: sth=%s\n", sth); }
 	
 	return imp_sth->rows;
 
@@ -1934,7 +1957,7 @@ int dbd_st_finish (sth, imp_sth)
 {
 	dTHR;
 	
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_finish\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_finish: sth=%s\n", sth); }
 	
 	if (DBIc_ACTIVE(imp_sth) && imp_sth->result) {
 		PQclear(imp_sth->result);
@@ -2097,7 +2120,7 @@ int dbd_st_STORE_attrib (sth, imp_sth, keysv, valuesv)
 	STRLEN vl;
 	char *value = SvPV(valuesv,vl);
 
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_STORE\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_STORE: sth=%s\n", sth); }
 	
 	if (17==kl && strEQ(key, "pg_server_prepare")) {
 		imp_sth->server_prepare = strEQ(value,"0") ? 0 : 1;
@@ -2128,10 +2151,9 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 	char *key = SvPV(keysv,kl);
 	int i, x, y, sz;
 	SV *retsv = Nullsv;
-	char *type_name;
 	sql_type_info_t *type_info;
 
-	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_FETCH\n"); }
+	if (dbis->debug >= 4) { PerlIO_printf(DBILOGFP, "dbd_st_FETCH: sth=%s\n", sth); }
 	
 	/* Some can be done before the execute */
 	if (15==kl && strEQ(key, "pg_prepare_name")) {
@@ -2152,11 +2174,11 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 		for (i=0,currph=imp_sth->ph; NULL != currph; currph=currph->nextph,i++) {
 			if (NULL == currph->value) {
 				hv_store_ent(pvhv, 3==imp_sth->placeholder_type ? newSVpv(currph->fooname,0) : 
-										 newSViv(i+1), Nullsv, i);
+										 newSViv(i+1), Nullsv, (unsigned)i);
 			}
 			else {
 				hv_store_ent(pvhv, 3==imp_sth->placeholder_type ? newSVpv(currph->fooname,0) : 
-										 newSViv(i+1), newSVpv(currph->value,0),i);
+										 newSViv(i+1), newSVpv(currph->value,0),(unsigned)i);
 			}
 		}
 		retsv = newRV_noinc((SV*)pvhv);
@@ -2180,7 +2202,7 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 		AV *av = newAV();
 		retsv = newRV(sv_2mortal((SV*)av));
 		while(--i >= 0) {
-			type_info = pg_type_data(PQftype(imp_sth->result, i));
+			type_info = pg_type_data((int)PQftype(imp_sth->result, i));
 			av_store(av, i, newSViv( type_info ? type_info->type.sql : 0 ) );
 		}
 	}
@@ -2221,7 +2243,7 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 	else if (8==kl && strEQ(key, "NULLABLE")) {
 		AV *av = newAV();
 		PGresult *result;
-		PGTransactionStatusType status;
+		int status = -1;
 		D_imp_dbh_from_sth;
 		char *statement;
 		int nullable; /* 0 = not nullable, 1 = nullable 2 = unknown */
@@ -2239,7 +2261,8 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 				sprintf(statement, "SELECT attnotnull FROM pg_catalog.pg_attribute WHERE attrelid=%d AND attnum=%d", x, y);
 				statement[strlen(statement)]='\0';
 				result = PQexec(imp_dbh->conn, statement);
-				status = imp_sth->result ? PQresultStatus(imp_sth->result) : -1;
+				if (result)
+					status = PQresultStatus(result);
 				if (PGRES_TUPLES_OK == status && PQntuples(result)) {
 					switch(PQgetvalue(result,0,0)[0]) {
 					case 't':
@@ -2272,13 +2295,12 @@ SV * dbd_st_FETCH_attrib (sth, imp_sth, keysv)
 		AV *av = newAV();
 		retsv = newRV(sv_2mortal((SV*)av));
 		while(--i >= 0) {			
-			type_info = pg_type_data(PQftype(imp_sth->result,i));
-			type_name = (type_info) ? type_info->type_name : "unknown";
-			av_store(av, i, newSVpv(type_name, 0));			
+			type_info = pg_type_data((int)PQftype(imp_sth->result,i));
+			av_store(av, i, newSVpv(type_info ? type_info->type_name : "unkown", 0));
 		}
 	}
 	else if (13==kl && strEQ(key, "pg_oid_status")) {
-		retsv = newSViv(PQoidValue(imp_sth->result));
+		retsv = newSViv((int)PQoidValue(imp_sth->result));
 	}
 	else if (13==kl && strEQ(key, "pg_cmd_status")) {
 		retsv = newSVpv((char *)PQcmdStatus(imp_sth->result), 0);
@@ -2314,7 +2336,7 @@ pg_db_putline (dbh, buffer)
 		if (dbis->debug >= 4)
 			PerlIO_printf(DBILOGFP, "  dbdpg: PQputCopyData\n");
 
-		result = PQputCopyData(imp_dbh->conn, buffer, strlen(buffer));
+		result = PQputCopyData(imp_dbh->conn, buffer, (int)strlen(buffer));
 		if (-1 == result) {
 			pg_error(dbh, PQstatus(imp_dbh->conn), PQerrorMessage(imp_dbh->conn));
 			return 0;
@@ -2355,6 +2377,7 @@ pg_db_getline (dbh, buffer, length)
 		tempbuf = NULL; /* Make compilers happy */
 		return result;
 #else
+		length = 0; /* Make compilers happy */
 		result = PQgetCopyData(imp_dbh->conn, &tempbuf, 0);
 		if (-1 == result) {
 			*buffer = '\0';
@@ -2579,4 +2602,203 @@ int pg_db_release (dbh, imp_dbh, savepoint)
 	return 1;
 }
 
+
+/* ================================================================== */
+/* Large object functions */
+
+unsigned int pg_db_lo_creat (dbh, mode)
+		SV *dbh;
+		int mode;
+{
+	int status = -1;
+	D_imp_dbh(dbh);
+
+	/* If not autocommit, start a new transaction */
+	if (!imp_dbh->done_begin && DBDPG_FALSE == DBIc_has(imp_dbh, DBIcf_AutoCommit)) {
+		status = _result(imp_dbh, "begin");
+		if (PGRES_COMMAND_OK != status) {
+			pg_error(dbh, status, PQerrorMessage(imp_dbh->conn));
+			return (unsigned)-2;
+		}
+		imp_dbh->done_begin = 1;
+	}
+	return lo_creat(imp_dbh->conn, mode);
+}
+
+int pg_db_lo_open (dbh, lobjId, mode)
+		 SV *dbh;
+		 unsigned int lobjId;
+		 int mode;
+{
+	D_imp_dbh(dbh);
+	return lo_open(imp_dbh->conn, lobjId, mode);
+}
+
+int pg_db_lo_close (dbh, fd)
+		 SV *dbh;
+		 int fd;
+{
+	D_imp_dbh(dbh);
+	return lo_close(imp_dbh->conn, fd);
+}
+
+int pg_db_lo_read (dbh, fd, buf, len)
+		 SV *dbh;
+		 int fd;
+		 char *buf;
+		 unsigned int len;
+{
+	D_imp_dbh(dbh);
+	return lo_read(imp_dbh->conn, fd, buf, len);
+}
+
+
+int pg_db_lo_write (dbh, fd, buf, len)
+		 SV *dbh;
+		 int fd;
+		 char *buf;
+		 unsigned int len;
+{
+	D_imp_dbh(dbh);
+	return lo_write(imp_dbh->conn, fd, buf, len);
+}
+
+
+int pg_db_lo_lseek (dbh, fd, offset, whence)
+		 SV *dbh;
+		 int fd;
+		 int offset;
+		 int whence;
+{
+	D_imp_dbh(dbh);
+	return lo_lseek(imp_dbh->conn, fd, offset, whence);
+}
+
+
+int pg_db_lo_tell (dbh, fd)
+		SV *dbh;
+		int fd;
+{
+	D_imp_dbh(dbh);
+	return lo_tell(imp_dbh->conn, fd);
+}
+
+
+int pg_db_lo_unlink (dbh, lobjId)
+		 SV *dbh;
+		 unsigned int lobjId;
+{
+	D_imp_dbh(dbh);
+	return lo_unlink(imp_dbh->conn, lobjId);
+}
+
+
+unsigned int pg_db_lo_import (dbh, filename)
+		 SV *dbh;
+		 char *filename;
+{
+	D_imp_dbh(dbh);
+	return lo_import(imp_dbh->conn, filename);
+}
+
+
+int pg_db_lo_export (dbh, lobjId, filename)
+		 SV *dbh;
+		 unsigned int lobjId;
+		 char *filename;
+{
+	D_imp_dbh(dbh);
+	return lo_export(imp_dbh->conn, lobjId, filename);
+}
+
+int dbd_st_blob_read (sth, imp_sth, lobjId, offset, len, destrv, destoffset)
+		 SV *sth;
+		 imp_sth_t *imp_sth;
+		 int lobjId;
+		 long offset;
+		 long len;
+		 SV *destrv;
+		 long destoffset;
+{
+	D_imp_dbh_from_sth;
+	int ret, lobj_fd, nbytes;
+	STRLEN nread;
+	SV *bufsv;
+	char *tmp;
+	
+	if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_st_blob_read\n"); }
+	/* safety checks */
+	if (lobjId <= 0) {
+		pg_error(sth, -1, "dbd_st_blob_read: lobjId <= 0");
+		return 0;
+	}
+	if (offset < 0) {
+		pg_error(sth, -1, "dbd_st_blob_read: offset < 0");
+		return 0;
+	}
+	if (len < 0) {
+		pg_error(sth, -1, "dbd_st_blob_read: len < 0");
+		return 0;
+		}
+	if (! SvROK(destrv)) {
+		pg_error(sth, -1, "dbd_st_blob_read: destrv not a reference");
+		return 0;
+	}
+	if (destoffset < 0) {
+		pg_error(sth, -1, "dbd_st_blob_read: destoffset < 0");
+		return 0;
+	}
+	
+	/* dereference destination and ensure it's writable string */
+	bufsv = SvRV(destrv);
+	if (! destoffset) {
+		sv_setpvn(bufsv, "", 0);
+	}
+	
+	/* open large object */
+	lobj_fd = lo_open(imp_dbh->conn, (unsigned)lobjId, INV_READ);
+	if (lobj_fd < 0) {
+		pg_error(sth, -1, PQerrorMessage(imp_dbh->conn));
+		return 0;
+	}
+	
+	/* seek on large object */
+	if (offset > 0) {
+		ret = lo_lseek(imp_dbh->conn, lobj_fd, offset, SEEK_SET);
+		if (ret < 0) {
+			pg_error(sth, -1, PQerrorMessage(imp_dbh->conn));
+			return 0;
+		}
+	}
+	
+	/* read from large object */
+	nread = 0;
+	SvGROW(bufsv, destoffset + nread + BUFSIZ + 1);
+	tmp = (SvPVX(bufsv)) + destoffset + nread;
+	while ((nbytes = lo_read(imp_dbh->conn, lobj_fd, tmp, BUFSIZ)) > 0) {
+		nread += nbytes;
+		/* break if user wants only a specified chunk */
+		if (len > 0 && (int)nread > len) {
+			nread = len;
+			break;
+		}
+		SvGROW(bufsv, destoffset + nread + BUFSIZ + 1);
+		tmp = (SvPVX(bufsv)) + destoffset + nread;
+	}
+	
+	/* terminate string */
+	SvCUR_set(bufsv, destoffset + nread);
+	*SvEND(bufsv) = '\0';
+	
+	/* close large object */
+	ret = lo_close(imp_dbh->conn, lobj_fd);
+	if (ret < 0) {
+		pg_error(sth, -1, PQerrorMessage(imp_dbh->conn));
+		return 0;
+	}
+	
+	return nread;
+}
+
 /* end of dbdimp.c */
+
